#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 by Raymond Hettinger, Łukasz Langa
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Taken from Raymond Hettinger's recipe at:

http://code.activestate.com/recipes/576694-orderedset/
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import collections

KEY, PREV, NEXT = range(3)

class OrderedSet(collections.MutableSet):
    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]         # sentinel node for doubly linked list
        self.map = {}                   # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[PREV]
            curr[NEXT] = end[PREV] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[NEXT] = next
            next[PREV] = prev

    def __iter__(self):
        end = self.end
        curr = end[NEXT]
        while curr is not end:
            yield curr[KEY]
            curr = curr[NEXT]

    def __reversed__(self):
        end = self.end
        curr = end[PREV]
        while curr is not end:
            yield curr[KEY]
            curr = curr[PREV]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = next(reversed(self)) if last else next(iter(self))
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)

    def __del__(self):
        self.clear()                    # remove circular references



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 by Łukasz Langa
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""lck.cache.memoization
   ---------------------

   Implements a reusable memoization decorator. It is using a finite-size cache
   with pickled arguments as keys, to hold the outcome of a specific function
   call. When the decorated function is called again with the same arguments,
   the outcome is fetched from the cache instead of being recalculated again.

   The cache used maintains a list of *Least Recently Used* keys so that in
   case of overflow only the seemingly least important ones get deleted.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import cPickle as pickle
import sys
from time import time
from functools import wraps


def memoize(func=None, update_interval=300, max_size=256, skip_first=False,
    fast_updates=True):
    """Memoization decorator.

        :param update_interval: time in seconds after which the actual function
                                will be called again

        :param max_size: maximum buffer count for distinct memoize hashes for
                         the function. Can be set to 0 or ``None``. Be aware of
                         the possibly inordinate memory usage in that case

        :param skip_first: ``False`` by default; if ``True``, the first
                           argument to the actual function won't be added to
                           the memoize hash

        :param fast_updates: if ``True`` (the default), an optimized LRU
                             algorithm is used where all function invocations
                             except every Nth (where N ==sys.maxint) are much
                             faster but cache overflow is costly. In general,
                             having ``fast_updates`` set to ``True`` gives
                             a 15% performance boost when there are no cache
                             misses (the possible number of used argument
                             combinations for the decorated function is smaller
                             than the value of ``max_size``). If cache misses
                             exceed 50%, you might want to increase
                             ``max_size``. If that's not feasible, memoization
                             with ``fast_updates`` set to ``False`` will
                             perform faster."""

    # the decorator can be used with an argument as well as without any
    if func is None:
        def wrapper(f):
            return memoize(func=f,
                           update_interval=update_interval,
                           max_size=max_size,
                           skip_first=skip_first,
                           fast_updates=fast_updates)
        return wrapper

    cached_values = {}

    if fast_updates:
        lru_indices = {'CURRENT': 0}
    else:
        lru_list = []

    @wraps(func)
    def wrapper_standard(*args, **kwargs):
        if skip_first:
            key = pickle.dumps((args[1:], kwargs))
        else:
            key = pickle.dumps((args, kwargs))

        lru_current = lru_indices['CURRENT'] + 1
        lru_indices[key] = lru_current

        if key in cached_values:
            # get the buffered values and check whether they are up-to-date
            result, acquisition_time = cached_values[key]
            if update_interval and time() - acquisition_time > update_interval:
                del cached_values[key]

        if key not in cached_values:
            result = func(*args, **kwargs)
            acquisition_time = time()
            cached_values[key] = (result, acquisition_time)

            # clear the least recently used value if the maximum size
            # of the buffer is exceeded (max_size+1 because of the magic
            # 'CURRENT' key)
            if max_size and len(lru_indices) > max_size + 1:
                lru_key = min(lru_indices.iteritems(), key=lambda x: x[1])[0]
                del lru_indices[lru_key]
                del cached_values[lru_key]

        if lru_current == sys.maxint:
            # renumber indices to avoid hurting performance by using bigints
            lru_current = 0
            for key, _ in sorted(lru_indices.iteritems(), key=lambda x: x[1]):
                lru_indices[key] = lru_current
                lru_current += 1

        lru_indices['CURRENT'] = lru_current

        return result

    @wraps(func)
    def wrapper_legacy(*args, **kwargs):
        if skip_first:
            key = pickle.dumps((args[1:], kwargs))
        else:
            key = pickle.dumps((args, kwargs))

        if key in cached_values:
            # move argument to the end of the LRU list
            lru_list.append(lru_list.pop(lru_list.index(key)))

            # get the buffered values and check whether they are up-to-date
            result, acquisition_time = cached_values[key]
            if update_interval and time() - acquisition_time > update_interval:
                del cached_values[key]
                lru_list.remove(key)

        if key not in cached_values:
            cached_values[key] = (func(*args, **kwargs), time())
            lru_list.append(key)

            # clear the least recently used value if the maximum size
            # of the buffer is exceeded
            if max_size and len(lru_list) > max_size:
                del cached_values[lru_list.pop(0)]

        return cached_values[key][0]

    return wrapper_standard if fast_updates else wrapper_legacy



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 by Łukasz Langa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Synchronization tests
   ---------------------

   Tests use the ``py.test`` framework. Run as::

       $ easy_install -U py
       $ py.test
   """

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from threading import Thread, Lock, RLock
from time import time, sleep

from lck.concurrency import synchronized

SLEEP_AMOUNT=0.05 #seconds

def test_simple_synchronization(sleep_amount=SLEEP_AMOUNT):
    @synchronized
    def simple_synchro(result):
        for i in xrange(10):
            result.append(i)
            sleep(sleep_amount)

    class SimpleThread(Thread):
        def __init__(self, result):
            Thread.__init__(self)
            self.result = result
        def run(self):
            simple_synchro(self.result)

    result = []
    threads = [SimpleThread(result) for i in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert result == range(10) * 10


def test_shared_synchronization(sleep_amount=SLEEP_AMOUNT):
    lock = Lock()

    class SharedLockThread(Thread):
        def __init__(self, result):
            Thread.__init__(self)
            self.result = result

        @synchronized(lock=lock)
        def run(self):
            for i in xrange(10):
                self.result.append(i)
                sleep(sleep_amount)

    result = []
    threads = [SharedLockThread(result) for i in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert result == range(10) * 10


def test_reentrant_synchronization(sleep_amount=SLEEP_AMOUNT):
    lock = RLock()

    class ReentrantThread(Thread):
        def __init__(self, result):
            Thread.__init__(self)
            self.result = result

        @synchronized(lock=lock)
        def run(self, i=0):
            if i > 9:
                return

            self.result.append(i)
            sleep(sleep_amount)
            self.run(i + 1)

    result = []
    threads = [ReentrantThread(result) for i in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert result == range(10) * 10


def test_filebased_synchronization(sleep_amount=SLEEP_AMOUNT):
    class FilesystemLockThread(Thread):
        def __init__(self, result):
            Thread.__init__(self)
            self.result = result

        @synchronized(path='./test.lock')
        def run(self):
            for i in xrange(10):
                self.result.append(i)
                sleep(sleep_amount)

    result = []
    threads = [FilesystemLockThread(result) for i in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert result == range(10) * 10



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 by Łukasz Langa
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""lck.config
   ----------

   Configuration related utilities."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import Callable

from ConfigParser import SafeConfigParser, NoOptionError, NoSectionError


class FunkyConfigParser(SafeConfigParser):
    """Config parser that can handle get() fallbacks and getting/setting lists
    of strings (white-space separated)."""

    def __init__(self, *args, **kwargs):
        self.is_fresh = True
        SafeConfigParser.__init__(self, *args, **kwargs)

    def readfp(self, *args, **kwargs):
        self.is_fresh = False
        SafeConfigParser.readfp(self, *args, **kwargs)

    def write(self, *args, **kwargs):
        self.is_fresh = False
        SafeConfigParser.write(self, *args, **kwargs)

    def set(self, section, option, value):
        if isinstance(value, list):
            SafeConfigParser.set(self, section, option, " ".join(value))
        else:
            SafeConfigParser.set(self, section, option, str(value))

    def get(self, section, option, fallback=None, conv=str):
        try:
            raw_value = SafeConfigParser.get(self, section, option)
        except (NoOptionError, NoSectionError):
            if fallback is None:
                raise
            elif isinstance(fallback, Callable):
                return fallback(section, option)
            else:
                return fallback
        if conv == list:
            return raw_value.strip().split()
        elif conv == bool:
            if raw_value.lower() not in self._boolean_states:
                raise ValueError('Not a boolean: %s' % raw_value)
            return self._boolean_states[raw_value.lower()]
        else:
            return conv(raw_value)

    def getint(self, section, option, fallback=None):
        return self.get(section, option, fallback, int)

    def getfloat(self, section, option, fallback=None):
        return self.get(section, option, fallback, float)

    def getlist(self, section, option, fallback=None):
        return self.get(section, option, fallback, list)

    def getboolean(self, section, option, fallback=None):
        return self.get(section, option, fallback, bool)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 by Łukasz Langa
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""lck.concurrency.synchronization
   -------------------------------

   Implements a reusable Java-like synchronization decorator.
   It is using threading locks or filesystem-based locks to synchronize
   subsequent calls of the specified functions. The former kind of
   lock is reentrant, the latter is not.

   For filesystem-based locks the module is using Skip Montanaro's
   `lockfile <http://pypi.python.org/pypi/lockfile>`_ library,
   compatible with Windows and POSIX environments."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
from threading import RLock
from lockfile import FileLock
from functools import wraps

def synchronized(func=None, lock=None, path=None):
    """ Synchronization decorator.

        :param lock: the user can specify a concrete lock object to be used
               with this specific synchronization decorator. This is
               useful when a group of functions should be synchronized
               together.
        :param path: instead of using threading-based locking, file-based
               locks may be used instead. Beware, these are radically
               less performant than threading locks.
    """

    # the decarator can be used with an argument as well as without any
    if func is None:
        def wrapper(f):
            return synchronized(f, lock=lock)
        return wrapper

    if path is None:
        _lock = lock if lock else RLock()
    else:
        _lock = FileLock(path)

    @wraps(func)
    def wrapper(*args, **kwargs):
        with _lock:
            result = func(*args, **kwargs)
        return result

    return wrapper



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 by Łukasz Langa
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Memoization tests
   -----------------

   Tests use the ``py.test`` framework. Run as::

       $ easy_install -U py
       $ py.test
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from time import time, sleep

from lck.cache import memoize

def _update_interval_test(current_time):
    t = current_time()
    sleep(1)
    assert t == current_time()
    sleep(1)
    assert t == current_time()
    sleep(3)
    assert t != current_time()

def _max_size_test(current_time):
    t1 = current_time(1)
    sleep(1)
    t2 = current_time(2)

    sleep(1)
    assert t1 == current_time(1)
    assert t2 == current_time(2)

    t3 = current_time(3)
    sleep(1)
    t4 = current_time(4)

    sleep(1)
    assert t3 == current_time(3)
    assert t4 == current_time(4)
    assert t1 != current_time(1)
    assert t2 != current_time(2)

    sleep(1)
    assert t3 != current_time(3)
    assert t4 != current_time(4)

def test_memoization_update_interval():
    @memoize(fast_updates=False, update_interval=4)
    def current_time():
        return time()
    _update_interval_test(current_time)

def test_memoization_max_size():
    @memoize(fast_updates=False, max_size=2)
    def current_time(arg):
        return time()
    _max_size_test(current_time)

def test_memoization_fast_updates_update_interval():
    @memoize(fast_updates=True, update_interval=4)
    def current_time():
        return time()
    _update_interval_test(current_time)

def test_memoization_fast_updates_max_size():
    @memoize(fast_updates=True, max_size=2)
    def current_time(arg):
        return time()
    _max_size_test(current_time)



# -*- encoding: utf-8 -*-
# Copyright (C) 2011 by Łukasz Langa
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import sys
from setuptools import setup, find_packages

assert sys.version_info >= (2, 7), "Python 2.7+ required."

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as ld_file:
    long_description = ld_file.read()

from doc.conf import release

setup (
    name = 'lck.common',
    version = release,
    author = 'Lukasz Langa',
    author_email = 'lukasz@langa.pl',
    description = ("A library of various simple common routines that keep "
                   "being rewritten all over again in every project I'm "
                   "working on."),
    long_description = long_description,
    url = 'http://packages.python.org/lck.common/',
    keywords = '',
    platforms = ['any'],
    license = 'MIT',
    packages = find_packages('src'),
    include_package_data = True,
    package_dir = {'':'src'},
    namespace_packages = ['lck'],
    zip_safe = True,
    install_requires = [
        'lockfile',
    ],
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows :: Windows NT/2000',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 by Łukasz Langa
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import math

def rate(winner_rank=1000, loser_rank=1000, penalize_loser=True):
    rank_diff = winner_rank - loser_rank
    exp = rank_diff / -400
    odds = 1 / (1 + math.pow(10, exp))
    if winner_rank < 2100:
        k = 32
    elif 2100 <= winner_rank < 2400:
        k = 24
    else:
        k = 16
    new_winner_rank = round(winner_rank + k * (1 - odds))
    new_loser_rank = loser_rank
    if penalize_loser:
        new_rank_diff = new_winner_rank - winner_rank
        new_loser_rank -= new_rank_diff
    return new_winner_rank, max(new_loser_rank, 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 by Łukasz Langa
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""lck.crypto.cipher
   -----------------

   Based on http://code.activestate.com/recipes/496763/."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from base64 import b64encode, b64decode
import os
import random
import stat
import string
import tempfile

from lck.concurrency import synchronized

ALPHABET = string.letters + string.digits + string.punctuation


class Cipher(object):
    def __init__(self, key=None, path=None, create=True, cipher=None):
        """Using a provided `key` or `path` to a file with the key, create
        a Cipher capable of encrypting and decrypting bytes. `cipher` is
        the algorithm to be used.

        If `path` is provided and `create` is True, if the file under `path`
        is absent, it is created."""

        if key is None:
            if path is None:
                raise ValueError('Either `key` or `path` must be provided.')
            length = 32 if not cipher.key_size else cipher.key_size
            key = self.key_from_path(path, create, length)
        self._cipher = cipher.new(key)
        self.block = self._cipher.block_size

    @staticmethod
    @synchronized(path=os.path.join(tempfile.gettempdir(),
        'lck-crypto-cipher.lock'))
    def key_from_path(path, create=True, length=32):
        if not os.path.exists(path):
            if not create:
                raise IOError("File does not exist: {}".format(path))
            else:
                with open(path, "wb") as key_file:
                    key_file.write(''.join([random.choice(ALPHABET)
                        for i in range(length)]))
                    key_file.flush()
                    os.fchmod(key_file.fileno(), stat.S_IRUSR)
        with open(path, "rb") as key_file:
            return key_file.read()

    def encrypt(self, buffer, base64=False):
        assert isinstance(buffer, bytes), "Cannot encrypt {} instances. "\
            "Encode to bytes first.".format(buffer.__class__.__name__)
        result = self._cipher.encrypt(self._pad_buffer(buffer))
        if base64:
            result = b64encode(result)
        return result

    def decrypt(self, buffer, base64=False):
        assert isinstance(buffer, bytes), "Cannot decrypt {} instances. "\
            "Encode to bytes first.".format(buffer.__class__.__name__)
        if base64:
            buffer = b64decode(buffer)
        result = self._depad_buffer(self._cipher.decrypt(buffer))
        return result

    # the cipher needs self.block byte blocks to work with
    def _pad_buffer(self, buffer):
        pad_bytes = self.block - (len(buffer) % self.block)
        for i in range(pad_bytes - 1):
            buffer += chr(random.randrange(0, 256))
        # final padding byte; % by self.block to get the number of padding bytes
        bflag = random.randrange(6, 256 - self.block)
        bflag -= bflag % self.block - pad_bytes
        buffer += chr(bflag)
        return buffer

    def _depad_buffer(self, buffer):
        pad_bytes = ord(buffer[-1]) % self.block
        if not pad_bytes:
            pad_bytes = self.block
        return buffer[:-pad_bytes]



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 by Łukasz Langa
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""lck.files.finder
   ----------------

   Provides a function for searching for a file in sensible locations."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os


class CommonPathPrefixes(list):
    def __init__(self):
        """Instantiates common path prefixes used in search for specific file.
           For now these are: the current and parent directory, subdirectories
           of the current directory, possible home directories for the current
           user.

           ..note::
             We don't check the __file__ directory of any module because we
             believe data should not be mixed with code."""
        self.append('.')
        self.append('..')
        for entry in os.listdir('.'):
            if not os.path.isdir(entry):
                continue
            self.append(os.path.join('.', entry))
        if 'HOME' in os.environ:
            self.append(os.environ['HOME'])
        elif 'USER' in os.environ:
            self.append('/home/{USER}'.format(os.environ))
            self.append('/home/users/{USER}'.format(os.environ))
            self.append('/Users/{USER}'.format(os.environ))
        self.append('/etc')

    def generate(self, *args):
        """Returns a generator that yields existing paths for every permutation
           of the common path prefixes with paths in the specified list. Paths
           are permuted only when the one from the given list is not
           absolute."""

        for p in args:
            if not p:
                continue
            elif os.path.isabs(p):
                if os.path.exists(p):
                    yield p
                continue
            for prefix in self:
                generated_path = os.path.join(prefix, p)
                if os.path.exists(generated_path):
                    yield generated_path

common_path_prefixes = CommonPathPrefixes()

def finder(explicit_path, envvar=None, multiple_allowed=False):
    """Finds a specific file using explicitly given path (or given by an
       environment variable). The algorithm is as follows: for every given
       path from the args (explicitly given, environment variable, fallback)
       check whether the file exists. If it doesn't and the path is not
       absolute, search the working directory, its parent directory and all
       child directories, the current user's home directory and /etc.

       :param explicit_path: path explicitly given by the user, can be a single
                             entry or a sequence

       :param envvar: name of the environment variable where to look for the
                      path

       :param fallback: name of the file to check if everything else fails

       :param multiple_allowed: False by default. If True, the returned type is
                                a tuple with potentially many entries.

       :return: the real absolute path to the file. Raises IOError if no found.

       ..note::
         Works only on POSIX systems."""

    if not explicit_path:
        explicit_path = []
    elif explicit_path.__class__ in (list, tuple, set):
        # list copying for safety reasons (routine is side-effect free)
        explicit_path = list(explicit_path)
    else:
        explicit_path = [explicit_path]

    explicit_path.append(os.environ.get(envvar))

    # keep only given arguments
    existing_paths = common_path_prefixes.generate(*explicit_path)

    try:
        if multiple_allowed:
            existing_paths = list(existing_paths)
            if not existing_paths:
                raise StopIteration()
            return existing_paths
        else:
            return existing_paths.next()
    except StopIteration:
        raise IOError("File not found.")
